;;; emacs-panel.el --- gnome-panel-like functionality for Emacs

;; Copyright (C) 2011  Teodor Zlatanov

;; Author: Teodor Zlatanov <tzz@lifelogs.com>

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; example:

;; (require 'battery)
;; (require 'timeclock)

;; (emacs-panel-popup-add
;;  "status" :function
;;  (lambda ()
;;    (format "%s\n%s\n%s\n%s"
;;            (format-time-string "%H:%M %Y-%m-%d" (current-time))
;;            (battery-format battery-echo-area-format (funcall
;;                                                      battery-status-function))
;;            (timeclock-status-string)
;;            (shell-command-to-string "nmcli -p dev"))))

;;; --OR--

;; (emacs-panel-tile-add "myfeeds" :margin 3 :text ... :call ...)
;; (emacs-panel-tile-add 'mychat :padding 2 :update-function ... :call ...)

;; and then use it in a tiled 3x3 layout:

;; (emacs-panel-popup-add "status" :tiles '(nil "myfeeds" nil 'mychat)
;;                                 :background "black"
;;                                 :layout-manager '(tile :x 3 :y 3))


;(emacs-panel-popup-remove "status")

;;; Code:

(eval-when-compile (require 'cl))
(require 'timer)
(require 'bindat)

(defgroup emacs-panel nil
  "The Emacs panel."
  :version "24.1")

(defvar emacs-panel-popup-frame)

(defvar emacs-panel-popup-timer
  (run-at-time t 1 'emacs-panel-popup-runqueue))

(defcustom emacs-panel-popups nil
  "Alist of defined popup buffers
format: (NAME :function F :frame-params P ...)

The keys besides :function (called on a refresh)
and :frame-params (used for new frames) are: ... TODO ..."
  :group 'emacs-panel)

(defcustom emacs-panel-popup-interval 5
  "How often to refresh the popups, in seconds."
  :group 'emacs-panel
  :type 'integer)

(defvar emacs-panel-current-cycle 0)

(defun emacs-panel-popup-add (name &rest params)
  (unless (emacs-panel-popup-entry name)
    (add-to-list 'emacs-panel-popups `(,name ,@params))))

(defun emacs-panel-popup-remove (name)
  (setq emacs-panel-popups (remove* name emacs-panel-popups
                                    :key 'car :test 'equal)))

(defun emacs-panel-popup-entry (name)
  (cdr-safe (assoc name emacs-panel-popups)))

(defun emacs-panel-popup-runqueue ()
  (interactive)
  (incf emacs-panel-current-cycle)
  (when (zerop (mod emacs-panel-current-cycle emacs-panel-popup-interval))
    (loop for p in emacs-panel-popups
          do (emacs-panel-popup (car-safe p)))))

(defun emacs-panel-popup-strut-frame ()
  (let* ((width 200)
         (height 500)
         (ff (make-frame `((visibility . nil)
                           (width . 20)))))
    (x-change-window-property "_NET_WM_STRUT_PARTIAL"
                              `(,width 0 0 0 0 ,height 0 0 0 0 0 0) ff
                              "CARDINAL" 32 t)
    (x-change-window-property "_NET_WM_WINDOW_TYPE"
                              '("_NET_WM_WINDOW_TYPE_DOCK") ff
                              "ATOM" 32 t)
    (make-frame-visible ff)))

(defconst emacs-panel-MWM
  `((hints (functions . ,(ash 1 0))
           (decorations . ,(ash 1 1))
           (input-mode . ,(ash 1 2))
           (status . ,(ash 1 3)))
    (funcs (all . ,(ash 1 0))
           (resize . ,(ash 1 1))
           (move . ,(ash 1 2))
           (minimize . ,(ash 1 3))
           (maximize . ,(ash 1 4))
           (close . ,(ash 1 5)))
    (decors (all . ,(ash 1 0))
           (border . ,(ash 1 1))
           (resizeh . ,(ash 1 2))
           (title . ,(ash 1 3))
           (menu . ,(ash 1 4))
           (minimize . ,(ash 1 5))
           (maximize . ,(ash 1 6))))
  "WM hints for emacs-panel functions")

(defun emacs-panel-popup-special-frame (hints funcs decors &rest frame-props)
  (let ((ff (make-frame `((visibility . nil) ,@frame-props)))
        (special-data (make-list 5 0))
        (todo '(hints funcs decors))
        (pos 0))
    (while todo
      (let* ((sym (pop todo))
             (info (cdr (assq sym emacs-panel-MWM)))
             (v (symbol-value sym))
             (v (if (consp v) v (list v)))
             (sum (apply '+
                         (mapcar (lambda (s) (or (cdr (assq s info)) 0)) v))))
        (setf (nth pos special-data) sum)
        (incf pos)))
    (progn
      (message "Popping up frame with special properties %S"
               special-data)
      (x-change-window-property "_MOTIF_WM_HINTS" special-data ff
                                "_MOTIF_WM_HINTS" 32 t)
      (make-frame-visible ff)
      ff)))

;; To make a frame without decoration (gtk_window_decorated):
;; (emacs-panel-popup-special-frame 'decorations nil nil)
;; To make a frame without delete (gtk_window_deletable):
;; (emacs-panel-popup-special-frame 'functions '(close all) nil)

(defun emacs-panel-popup (name &optional hints funcs decors)
  (let* ((old-frame (selected-frame))
        (entry (emacs-panel-popup-entry name))
        (func (plist-get entry :function))
        (frame-params (plist-get entry :frame-params))
        (bufname (concat " *" name "*")))
    (save-excursion
      (with-current-buffer (get-buffer-create bufname)

        (make-local-variable 'emacs-panel-popup-frame)
        (if (and (boundp 'emacs-panel-popup-frame)
                 emacs-panel-popup-frame
                 (member emacs-panel-popup-frame (frame-list)))
            (select-frame emacs-panel-popup-frame)
          (let ((default-frame-alist `((minibuffer . nil)
                                       (width . 20)
                                       (border-width . 0)
                                       (menu-bar-lines . 0)
                                       (tool-bar-lines . 0)
                                       (unsplittable . t)
                                       (left-fringe . 0)
                                       ,@frame-params))
                (pop-up-frame-function (lambda nil
                                         (emacs-panel-popup-special-frame
                                          hints funcs decors
                                          pop-up-frame-alist))))
                (switch-to-buffer-other-frame bufname)
            (setq emacs-panel-popup-frame (selected-frame))))
        (when func
          (setq mode-line-format nil)
          (redraw-modeline)
          ;; TODO: optimize and improve
          (erase-buffer)
          (insert (funcall func)))))
    (select-frame old-frame)))

(defmacro emacs-panel-x-property (prop window &optional type vec)
  `(x-window-property ,prop nil ,(or type "AnyPropertyType") ,window nil ,vec))

(defmacro emacs-panel-x-property-nullsepstringarray (prop window &optional type)
  `(split-string (emacs-panel-x-property ,prop ,window ,type) "\0" t))

(defmacro emacs-panel-x-property-u32r (prop window &optional type n)
  `(let ((spec '((:v ,@(when n (list 'vec n)) u32r)))
         (bin (emacs-panel-x-property ,prop ,window ,type))
         v)
     (cdr-safe (assq :v (bindat-unpack spec bin)))))

(defun emacs-panel-wm-hints ()
  ;; ask the root window what window to query for the WM name
  (let* ((nqid (emacs-panel-x-property-u32r "_NET_SUPPORTING_WM_CHECK" 0))
         (name (emacs-panel-x-property "_NET_WM_NAME" nqid)))
    `((name ,name)
      (desktop-names
       ,@(emacs-panel-x-property-nullsepstringarray "_NET_DESKTOP_NAMES" 0))
      (active-window
       ,(emacs-panel-x-property-u32r "_NET_ACTIVE_WINDOW" 0))
      (desktop-count
       ,(emacs-panel-x-property-u32r "_NET_NUMBER_OF_DESKTOPS" 0))
      (desktop-geometry
       ,@(let ((d (emacs-panel-x-property-u32r "_NET_DESKTOP_GEOMETRY"
                                               0 nil 2)))
          (list (aref d 0) (aref d 1))))
      (current-desktop
       ,(emacs-panel-x-property-u32r "_NET_CURRENT_DESKTOP" 0)))))

(defmacro emacs-panel-wm-get-desktop (&optional frame)
  `(emacs-panel-x-property-u32r
    "_NET_WM_DESKTOP"
    (string-to-number (cdr (assq 'outer-window-id (frame-parameters ,frame))))))

(defmacro emacs-panel-wm-set-active-desktop (desktop)
  `(x-send-client-message nil 0 nil "_NET_CURRENT_DESKTOP" 32 '(,desktop 0)))

(defun emacs-panel-unload-function ()
  (cancel-timer emacs-panel-popup-timer))

(provide 'emacs-panel)
;;; emacs-panel.el ends here
